subroutine uvdebug_pointers(name,text,huv)
  use image_def
  use gbl_message
  use imager_interfaces, only : map_message
  !! IMAGER - Debug the Mosaic offset pointers
  character(len=*), intent(in) :: name    !! Caller name
  character(len=*), intent(in) :: text    !! Buffers name
  type (gildas), intent(in) :: huv        !! UV header
  !
  character(len=80) :: chain
  !
  write(chain,'(A,I6,I6,A,I6,I6,A,I4)') ', X (Point) ',&
    & huv%gil%column_pointer(code_uvt_xoff), &
    & huv%gil%column_pointer(code_uvt_yoff), &
    & ',  L (phase) ', &
    & huv%gil%column_pointer(code_uvt_loff), &
    & huv%gil%column_pointer(code_uvt_moff), &
    & ',  ID (Source) ', &
    & huv%gil%column_pointer(code_uvt_id)
  call map_message(seve%d,name,text//chain,1)
end subroutine uvdebug_pointers
!
subroutine uv_map_comm(line,comm,error)
  use clean_arrays
  use clean_default
  use gbl_message
  use gkernel_interfaces
  use imager_interfaces, except_this => uv_map_comm
  !---------------------------------------------------------------------
  ! @ private
  !*
  ! IMAGER -- Support for commands
  ! UV_MAP or UV_RESTORE /FIELDS /RANGE /TRUNCATE  /CONTINUUM /INDEX  
  !   or  
  ! UV_SELF /RANGE /RESTORE
  !!
  !---------------------------------------------------------------------
  character(len=*), intent(in) :: comm     !! Caller name
  character(len=*), intent(inout) :: line  !! Command line
  logical, intent(inout) :: error          !! Logical error flag
  !
  ! Constants
  real(kind=8), parameter :: pi=3.14159265358979323846d0
  real(kind=4), parameter :: sec_to_rad=pi/3600.0/180.0
  character(len=1), parameter :: question_mark='?'
  ! ' UV_MAP', '/TRUNCATE',  '/FIELDS', '/RANGE', '/CONTINUUM', '/INDEX'
  integer, parameter :: o_trunc=1
  integer, parameter :: o_field=2
  integer, parameter :: o_range=3
  integer, parameter :: o_cont=4
  integer, parameter :: o_index=5
  ! ' UV_SELF', '/RANGE', '/RESTORE'
  integer, parameter :: opt_rang=1  ! /RANGE Option for UV_SELF
  integer, parameter :: opt_rest=2  ! /RESTORE Option for UV_SELF
  ! Memories ...
  logical, save :: use_cont=.false.
  logical, save :: do_self=.false.
  !
  ! Local ---
  logical :: use_self, do_cct, do_init
  logical :: true_mosaic   ! Distiniguish Mosaics from Single pointings
  integer :: saved_nfields ! Also used to distinguish Mosaic mode
  integer :: nline, ier, ifi
  character(len=2) :: code
  character(len=16) :: acomm
  type(gildas) :: htmp
  type(gildas) :: htmp2
  logical :: err, goon
  character(len=80) :: chain, string
  real :: offmax
  integer :: debug
  !
  ! Code ----
  call imager_tree('UV_MAP_COMM',.false.)
  debug = 0
  call sic_get_inte('DEBUG_UVMAP',debug,error)
  error = .false.
  !
  ! Check command line syntax
  call map_query (line,comm,themap,goon,error) 
  if ((.not.goon).or.error) return
  !
  err = .false.
  use_self = comm.eq.'UV_SELF'      ! Self Calibration image !
  if (use_self) then
    if (huv%gil%column_pointer(code_uvt_if).ne.0) then
      call map_message(seve%e,comm, &
      & 'This UV table (probably a result of UV_CONTINUUM) is not suited for self-calibration')
      error = .true.
      return
    endif
  endif
  do_init = hself%loca%size.eq.0 .or. sic_present(opt_rang,0)
  !
  true_mosaic = themap%nfields.ne.0
  if (true_mosaic) then
    offmax = 0
    do ifi = 1,abs(themap%nfields)
      offmax = max(offmax,(themap%offxy(1,ifi)**2+themap%offxy(2,ifi)**2))
    enddo
    offmax = sqrt(offmax)/sec_to_rad
    if (offmax.lt.mosaic_tolerance) then
      write(chain,'(A,F8.4,A,F8.4,A)') 'Treating tiny mosaic as a single field (Size ', &
        & offmax,' < Tolerance ',mosaic_tolerance,') '
      call map_message(seve%w,comm,chain)
      true_mosaic = .false.
    endif
  endif
  !
  if (true_mosaic) then
    saved_nfields = themap%nfields
    !
    ! Mosaic case
    if (use_self) then
      call map_message(seve%w,'UV_SELF','not fully tested for Mosaics')
      error = .false.
      string = defmap_center
      if (.not.do_self) do_weig = .true.
      !
      ! Self-calibrated data
      do_cct = sic_present(opt_rest,0) ! /RESTORE option
      if (.not.do_cct) then
        if (do_init) then
          if (hself%loca%size.ne.0) call uvdebug_pointers(comm,'Hself' ,hself)
          call init_selfcal('UV_SELF',line,error)
          if (error) return
          call uvdebug_pointers(comm,'Hself ',hself)
          !! call uv_dump_buffers('UV_SELF *')
          line = 'UV_SELF'
          nline = len_trim(line)
          call sic_analyse(acomm,line,nline,error)
        endif
      else
        if (hself%loca%size.eq.0) then
          call map_message(seve%e,comm,'/RESTORE  Self calibrated UV data not defined')
          error = .true.
        else if (sic_narg(opt_rang).ge.0) then
          call map_message(seve%e,comm,'/RESTORE  conflicts with /RANGE')
          error = .true.
        endif
      endif
      if (error) return
      call uvdebug_pointers(comm,'Hself ',hself)
      call uvdebug_pointers(comm,'Hraw  ',hraw)
      call gildas_null(htmp,type='UVT')
      call uv_buffer_finduv(code)
      call gdf_copy_header(huv,htmp,error)
      call gdf_copy_header(hself,huv,error)
      call uvdebug_pointers(comm,'Huv    ',huv)
      duv => duvself
      ! ! call uv_dump_buffers('SELFCAL')
      !
      if (.not.do_cct) then ! UV_SELF only case
        if (huv%gil%column_pointer(code_uvt_xoff).ne.0) then
          call map_message(seve%i,'UV_MOSAIC','UV data is a Pointing Mosaic',1)
        else if (huv%gil%column_pointer(code_uvt_loff).ne.0) then
          call map_message(seve%i,'UV_MOSAIC','UV data is a Phase Mosaic',1)
        else if (huv%gil%column_pointer(code_uvt_id).ne.0) then
          call map_message(seve%i,'UV_MOSAIC','UV data is a Mosaic with field ID',1)
        else
          call map_message(seve%w,'UV_MOSAIC','UV data is a Mosaic of unknown type',1)
        endif
        !
        ! Shift and Convert to Pointing offsets only (common Phase center) if needed
        if ((huv%gil%column_pointer(code_uvt_id).eq.0).and.(do_init)) then
          ! The question is which UV data is being shifted ? 
          ! The RAW one ?  The Line one ? The compressed one ? 
          call uv_shift_mosaic(line,'UV_MOSAIC',error)  ! Convert to Pointing offsets
          if (error) then
            themap%nfields = saved_nfields
            return
          endif
          !
          ! Copy back headers since Column types may have changed
          call gdf_copy_header(huv,hself,error)
        endif
        !
        ! This should be done on the HRAW data also if different from the HSELF one
        if (hraw%gil%nchan.ne.1) then
          call map_message(seve%w,comm,'Shifting RAW mosaic also ',1)
          call gildas_null(htmp2,type='UVT')
          call gdf_copy_header(huv, htmp2, error)
          call gdf_copy_header(hraw, huv, error)
          duv => duvraw ! Or calib ????
          call uv_shift_mosaic(line,'UV_MOSAIC',error)  ! Convert to Pointing offsets
          call gdf_copy_header(huv,hraw,error)
          call gdf_copy_header(htmp2,huv,error)
          duv => duvself
        endif
        ! ! call uv_dump_buffers('DUMP')
        call uvdebug_pointers(comm,'Huv ',huv)
        call mosaic_uvmap('UV_MOSAIC',line,error)     ! Image
        do_cont = .false.  ! Next CLEAN will be no CONTINUUM
        !
      else ! UV_SELF /RESTORE case
        !
        ! Here, MOSAIC_RESTORE should work because the UV table is already sorted...
        call map_message(seve%w,comm,'UV data is a Mosaic, under test...')
        !   The UV table is a priori already sorted & shifted
        !        call uv_shift_mosaic(line,'UV_MOSAIC',error)  ! Convert to Pointing offsets
        !        if (error) return
        call mosaic_restore(line,error)
      endif
      defmap_center = string          ! Restore MAP_CENTER
      call uv_buffer_resetuv(code)
      ! ! call gdf_print_header(huv)
      ! ! Print *,'UVSELF Header '
      ! !    call uv_dump_buffers('SG_MAP')
      call gdf_copy_header(htmp,huv,err)
      call uvdebug_pointers(comm,'Hself ',hself)
      !
      themap%nfields = saved_nfields ! Restore the initial status of UV data
      if (error) return
      do_self = .true.
      !
      ! End UV_SELF MOSAIC case
      !
    else
      !
      ! Normal Mosaic imaging: UV_MAP or UV_RESTORE
      if (comm.eq.'UV_MAP') then
        !
        ! UV_MAP only
        if (huv%gil%column_pointer(code_uvt_xoff).ne.0) then
          call map_message(seve%i,'UV_MOSAIC','UV data is a Pointing Mosaic',1)
        else if (huv%gil%column_pointer(code_uvt_loff).ne.0) then
          call map_message(seve%i,'UV_MOSAIC','UV data is a Phase Mosaic',1)
        else if (huv%gil%column_pointer(code_uvt_id).ne.0) then
          call map_message(seve%i,'UV_MOSAIC','UV data is a Mosaic with field ID',1)
        else
          call map_message(seve%w,'UV_MOSAIC','UV data is a Mosaic of unknown type',1)
        endif
        !
        ! Shift and Convert to Pointing offsets only (common Phase center) if needed
        if (huv%gil%column_pointer(code_uvt_id).eq.0) then
              if (debug.gt.0) call uv_dump_buffers('UV_MOSAIC - SHIFT')
          call uv_shift_mosaic(line,'UV_MOSAIC',error)  ! Convert to Pointing offsets
          if (error) return
        endif
        !
        use_cont = sic_present(o_cont,0)
        if (use_cont) then
          call map_message(seve%i,comm,'/CONT option still experimental for Mosaics')
          ! Select Continuum data set (UV_MAP /CONT or UV_RESTORE after a UV_MAP /CONT)
              if (debug.gt.0) then
                call uv_dump_buffers('UV_MOSAIC - RAW')
                call uv_buffer_finduv(code)
                Print *,'Code ',code
              endif
          call switch_uv_cont(line,comm,.false.,code,htmp,error)
              if (debug.gt.0) then
                Print *,'Code ',code,' R ',size(duvr,1), size(duvr,2), associated(duvr)
                Print *,'Code ',code,' S ',size(duvs,1), size(duvs,2), associated(duvs)
              endif
          if (error) return
          !
              if (debug.gt.0)  call uv_dump_buffers('UV_MOSAIC - CONT')
          call mosaic_uvmap('UV_MOSAIC',line,error)     ! Image
              if (debug.gt.0)  call uv_dump_buffers('UV_MOSAIC - LINE')
          call uv_buffer_resetuv(code) 
          call gdf_copy_header(htmp,huv,err)
              if (debug.gt.0) then
                Print *,'Code ',code,' R ',size(duvr,1), size(duvr,2), associated(duvr)
                Print *,'Code ',code,' S ',size(duvs,1), size(duvs,2), associated(duvs)
                call uv_dump_buffers('UV_MOSAIC - ENDCONT')
              endif
          !
          ! Patch for buffer handling when sorting Mosaic (Only the Continuum is sorted)
          if (.not.associated(duv)) then
            if (code.eq.'r'.or.code.eq.'s') code = 'i'
            call uv_buffer_resetuv(code)
          endif
          if (error) return          
          do_cont = .true.   ! Next CLEAN will be CONTINUUM
        else
          expand_cont = 1
          call mosaic_uvmap('UV_MOSAIC',line,error)     ! Image          
          do_cont = .false.  ! Next CLEAN will be no CONTINUUM
        endif
      else
        !
        ! Here this is UV_RESTORE
        if (use_cont) then    ! Remembered from last UV_MAP command
          call uv_shift_mosaic(line,'UV_MOSAIC',error)  ! Convert to Pointing offsets
          if (error) return
          call map_message(seve%w,comm,'Cannot yet restore Continuum Mosaic')
          ! Select Continuum data set (UV_MAP /CONT or UV_RESTORE after a UV_MAP /CONT)
          call switch_uv_cont(line,comm,.false.,code,htmp,error)
          if (error) return
          !
          call mosaic_restore(line,error)     ! Image
          call uv_buffer_resetuv(code) 
          call gdf_copy_header(htmp,huv,err)
          if (error) return          
          !
          ! There can be a size mis-match here ...
          if (any(hcont%gil%dim(1:2).ne.hsky%gil%dim(1:2))) then
            write(chain,'(A,I0,A,I0,A)') 'Re-sizing continuum image to [', &
              &   hsky%gil%dim(1),',',hsky%gil%dim(2),']'
            call map_message(seve%w,comm,chain)
            deallocate(dcont,stat=ier)
            deallocate(dcont,stat=ier)
            allocate(dcont(hsky%gil%dim(1),hsky%gil%dim(2)),stat=ier)
            call gdf_copy_header(hsky,hcont,error)
            hcont%gil%ndim = 2
            hcont%loca%size = hcont%gil%dim(1)*hcont%gil%dim(2)
            call sic_delvariable('CONTINUUM',.false.,error)
            call sic_mapgildas ('CONTINUUM',hcont,error,dcont)
          endif          
          !
          dcont(:,:) = dsky(:,:,1) ! Update CONTINUUM image too
          hcont%gil%minloc = hsky%gil%minloc
          hcont%gil%maxloc = hsky%gil%maxloc
          hcont%gil%rmin = hsky%gil%rmin
          hcont%gil%rmax = hsky%gil%rmax
        else
          call map_message(seve%w,comm,'UV data is a Mosaic, under test...')
          call uv_shift_mosaic(line,'UV_MOSAIC',error)  ! Convert to Pointing offsets
          if (error) return
          call mosaic_restore(line,error)      
        endif
      endif
    endif
    !
    ! End MOSAIC case
  else  !  if (map_version.eq.0) then
    !
    call sub_mosaic('OFF',error)
    !
    ! Single-Field, New version
    if (use_self) then
      !
      string = defmap_center
      if (.not.do_self) do_weig = .true.
      !
      ! Self-calibrated data 
      do_cct = sic_present(opt_rest,0) ! /RESTORE option
      if (.not.do_cct) then
        if (hself%loca%size.eq.0 .or. sic_present(opt_rang,0)) then
          call init_selfcal('UV_SELF',line,error)
          if (error) return
          ! ! call uv_dump_buffers('UV_SELF *')
          line = 'UV_SELF'
          nline = len_trim(line)
          call sic_analyse(acomm,line,nline,error)
        endif
      else 
        if (hself%loca%size.eq.0) then
          call map_message(seve%e,comm,'/RESTORE  Self calibrated UV data not defined')
          error = .true.
        else if (sic_narg(opt_rang).ge.0) then
          call map_message(seve%e,comm,'/RESTORE  conflicts with /RANGE')
          error = .true.
        endif
      endif
      if (error) return  
      call gildas_null(htmp,type='UVT')
      call uv_buffer_finduv(code) 
      ! !    call uv_dump_buffers('SELFCAL')  !!
      call gdf_copy_header(huv,htmp,error)
      !
      call gdf_copy_header(hself,huv,error)
      duv => duvself
      !
      ! Here, SG_MAP works because the UV table is already sorted...
      call sg_map(comm,line,the_method,huv,do_cct,error)
      defmap_center = string          ! Restore MAP_CENTER
      call uv_buffer_resetuv(code) 
      ! !    call uv_dump_buffers('SG_MAP')
      call gdf_copy_header(htmp,huv,err)
      if (error) return
      do_self = .true.
    else
      ! Normal UV_MAP or UV_RESTORE
      string = defmap_center
      do_self = .false.
      !
      ! Select current data set or possibly its Continuum sibbling 
      do_cct = comm.eq.'UV_RESTORE'
      ! In a UV_RESTORE command, USE_CONT is remembered from last UV_MAP command
      ! In a UV_MAP command, USE_CONT is defined by the /CONT option
      if (.not.do_cct) use_cont = sic_present(o_cont,0)
      !
      if (use_cont) then
        !
        ! Select Continuum data set (UV_MAP /CONT or UV_RESTORE after a UV_MAP /CONT)
        call switch_uv_cont(line,comm,do_cct,code,htmp,error)
        defmap_center = string
        if (error) return
        !
        ! Perform the imaging
        if (do_cct) defmap_center = ' '   ! The Continuum UV data has already been shifted
        call sg_map(comm,line,the_method,huv,do_cct,error)
        defmap_center = string
        !
        ! Reset HUV buffer
        call uv_buffer_resetuv(code) 
        call gdf_copy_header(htmp,huv,err)
        if (error) return
        !
        if (do_cct) then
          ! There can be a size mis-match here ...
          if (any(hcont%gil%dim(1:2).ne.hclean%gil%dim(1:2))) then
            write(chain,'(A,I0,A,I0,A)') 'Re-sizing continuum image to [', &
              &   hclean%gil%dim(1),',',hclean%gil%dim(2),']'
            call map_message(seve%w,comm,chain)
            deallocate(dcont,stat=ier)
            deallocate(dcont,stat=ier)
            allocate(dcont(hclean%gil%dim(1),hclean%gil%dim(2)),stat=ier)
            call gdf_copy_header(hclean,hcont,error)
            hcont%gil%ndim = 2
            hcont%loca%size = hcont%gil%dim(1)*hcont%gil%dim(2)
            call sic_delvariable('CONTINUUM',.false.,error)
            call sic_mapgildas ('CONTINUUM',hcont,error,dcont)
          endif          
          !
          dcont(:,:) = dclean(:,:,1) ! Update CONTINUUM image too
          hcont%gil%minloc = hclean%gil%minloc
          hcont%gil%maxloc = hclean%gil%maxloc
          hcont%gil%rmin = hclean%gil%rmin
          hcont%gil%rmax = hclean%gil%rmax
        else
          do_cont = .true.  ! Next CLEAN, there will be a CONTINUUM image
        endif
      else
        ! Select the data set. Can be UV_DATA, UV_MODEL, or UV_RESIDUAL
        !
        if (current_uvdata.eq.'DATA_UV') then
          if (do_cct) defmap_center = ' '
          call sg_map(comm,line,the_method,huv,do_cct,error) 
          defmap_center = string
        else                
          ! We may have an issue about data sorting here, if
          ! the data is to be rotated. Shifting only cause problem
          ! if the data has never been sorted...
          !
          call gildas_null(htmp,type='UVT')
          call uv_buffer_finduv(code) 
          call gdf_copy_header(huv,htmp,error)
          !
          call uvdata_select(comm,error)
          if (error) return
          !
          ! Now the Imaging Proper
          if (do_cct) defmap_center = ' '
          call sg_map(comm,line,the_method,huv,do_cct,error)   
          defmap_center = string
          call uv_buffer_resetuv(code) 
          call gdf_copy_header(htmp,huv,err)
          if (error) return
          !
        endif
        do_cont = .false.  ! Next CLEAN will be no CONTINUUM
      endif
      if (error) return
    endif
    !
    if (.not.do_cct) call no_mosaic(comm)
    !
    ! End of Single-Field section
    !
  endif
  !
  if (comm.eq.'UV_RESTORE') then
    if (true_mosaic) then
      last_shown = 'SKY'
    else
      last_shown = 'CLEAN'
    endif
  else
    last_shown = 'DIRTY'
  endif
  call check_view(1,last_shown)  ! Force VIEW re-computation for this
  !
  call imager_tree('UV_MAP_COMM',.true.)
end subroutine uv_map_comm
!
subroutine no_mosaic(rname)
  use clean_def
  use clean_default
  use clean_arrays
  use gbl_message
  use gkernel_interfaces
  use imager_interfaces, only : map_message
  !---------------------------------------------------------------------  !
  ! @ private
  !*
  ! IMAGER -- Switch back prompt to IMAGER>
  !!
  !---------------------------------------------------------------------  !
  character(len=*), intent(in) :: rname  !! Caller name
  !
  logical :: error
  !
  if (user_method%mosaic) then
    call map_message(seve%i,rname,'Switching to NORMAL mode')
    call gprompt_set('IMAGER')
    user_method%trunca = 0.0
    call sic_delvariable('PRIMARY',.false.,error)
    hprim%gil%dim(1) = 1
    user_method%mosaic = .false.
  endif
end subroutine no_mosaic
!
subroutine init_selfcal(rname,line,error)
  use clean_arrays
  use clean_default
  use gbl_message
  use gkernel_interfaces
  use imager_interfaces, except_this => init_selfcal
  !---------------------------------------------------------------------
  ! @ private
  !*
  ! IMAGER -- Support for command 
  ! UV_SELF /RANGE [Min Max Type]
  !
  ! Create the Self-Calibration UV tables from the current one
  ! by averaging the specified range of channels.  
  !     a)  Hcalib and Hraw : the continuum compressed line UV data   
  !     b)  Hself : the Continuum UV data  
  !!
  !---------------------------------------------------------------------
  character(len=*), intent(in) :: rname     !! Caller name
  character(len=*), intent(inout) :: line   !! Command line
  logical, intent(inout) :: error           !! Logical error flag
  !
  ! Constants
  integer, parameter :: opt_rang=1
  integer(kind=4), parameter :: mtype=3
  integer, parameter :: mranges=50
  character(len=12) :: types(mtype),mytype
  data types /'CHANNEL','VELOCITY','FREQUENCY'/
  !
  ! Local ---
  integer :: SELF_COMPRESS    ! Test variable
  logical :: SELF_DEBUG=.false.
  character(len=12) :: comm='UV_COMPRESS '
  integer :: ni,nt,nu,nv,nchan,ncomp
  integer :: channels(3)
  integer, allocatable :: chflag(:)
  !
  integer :: itype
  character(len=12) :: ctype
  integer :: ndim, ier
  real(8) :: drange(mranges)
  integer :: nc(mranges),numchan,ntmp,i
  !
  real :: cpu0, cpu1
  real :: uvmax, uvmin
  integer :: nline, n, narg
  logical :: sorted, shift, needed
  real(kind=8) :: new(3)
  character(len=80) :: chain, acomm, string
  integer, allocatable :: goodch(:)
  !
  ! Code ----
  SELF_COMPRESS = 0
  call sic_get_inte('SELF_COMPRESS',SELF_COMPRESS,error)
  call sic_get_logi('SELF_DEBUG',self_debug,error)
  error = .false.
  !
  ! Code
  call map_message(seve%i,rname,'Initializing the "selfcal" UV data set',1)
  !
  ! Erase current Self-Calibration UV tables
  if (associated(duvself)) then
    if (associated(duvself,duvcalib)) then
      nullify(duvself)
    else
      deallocate(duvself)
    endif
  endif
  if (allocated(duvcalib)) deallocate(duvcalib)
  if (allocated(duvraw)) deallocate(duvraw)
  call sic_delvariable('UVRAW',.false.,error) 
  call sic_delvariable('UVCALIB',.false.,error) 
  call sic_delvariable('UVSELF',.false.,error) 
  error = .false.
  !
  ! Decode the command line arguments
  numchan = max(sic_narg(opt_rang),0)
  nc = 0
  if (mod(numchan,2).eq.0) then
    !  Default CTYPE is CHANNEL
    mytype = 'CHANNEL'
  else if (numchan.eq.1) then
    call sic_ch(line,opt_rang,numchan,ctype,ndim,.true.,error)
    if (ctype.ne.'*') then
      call map_message(seve%e,rname,'/RANGE option syntax error')
      error = .true.
      return
    endif
    numchan = 0
  else
    call sic_ke(line,opt_rang,numchan,ctype,ndim,.true.,error)
    if (error) return
    call sic_ambigs(rname,ctype,mytype,itype,types,mtype,error)
    if (error)  return
    numchan = numchan-1
  endif
  !
  do i=1,numchan
    call sic_r8(line,opt_rang,i,drange(i),.true.,error)
    if (error) return
  enddo
  !
  ! Then, shift, rotate and sort the initial UV Table
  narg = sic_narg(0)
  string = defmap_center
  n = len_trim(string)
  !
  if (narg.eq.0) then
    if (n.ne.0) then
      line = 'CALIBRATE'//char(92)//trim(rname)//' '//string(1:n)
      nline = len_trim(line)
      call sic_analyse(acomm,line,nline,error)
      if (error) return
    endif
  endif
  call map_center(line,rname,huv,shift,new,error)
  if (error) return
  !
  call gag_cpu(cpu0)
  needed = themap%uniform(2).ne.0
  call uv_sort (huv,duv,error,sorted,shift,new,uvmax,uvmin,needed)
  if (error) return
  if (.not.sorted) then
    ! Redefine SIC variables (mandatory)
    call map_uvgildas ('UV',huv,error,duv)
  endif
  call gag_cpu(cpu1)
  write(chain,'(A,F9.2)') 'Finished sorting ',cpu1-cpu0
  call map_message(seve%i,rname,chain)
  !
  ! Now, define the Raw UV table
  call gildas_null(hraw,type='UVT')
  call gildas_null(hcalib,type='UVT')
  call gildas_null(hself,type='UVT')
  !
  hcalib%gil%nchan = 0 ! No known channel in Calib
  !
  IF (SELF_COMPRESS.eq.0) THEN
    !
    ! Here we use the UV_AVERAGE method - Single channel output
    huv%r2d => duv
    ndim = huv%gil%nlead+3+huv%gil%ntrail
    allocate(duvraw(ndim,huv%gil%dim(2)),duvcalib(ndim,huv%gil%dim(2)),stat=ier)
    if (ier.ne.0) then
      call map_message(seve%e,rname,'Self Calibration UV table allocation error')
      error = .true.
      return
    endif
    hraw%r2d => duvraw
    !
    ! At this stage, we should reset MAP_CENTER...
    acomm = 'LET MAP_CENTER " "'
    call exec_command(acomm,error)
    !
    ! Convert range to channels
    allocate(goodch(huv%gil%nchan))
    goodch = 1
    nc = 0
    if (numchan.eq.0) then
      numchan = 2
      nc(1) = 1
      nc(2) = huv%gil%nchan 
      !
      ! Could use here the same limit as defined by UV_CHECK BEAM
      call channel_restrict(rname,nc(1),nc(2),error)
      error = .false.
      if (allocated(dchanflag)) then
        ! DCHANFLAG can be allocated to a different size then the current HUV data (e.g. after UV_CONTINUUM)
        if (huv%gil%nchan.eq.size(dchanflag)) goodch(:)  = dchanflag  ! Filter line if needed
      endif
    else if (mytype.eq.'CHANNEL') then
      nc(1:numchan) = nint(drange(1:numchan))
    else if (mytype.eq.'VELOCITY') then
      ! drange = nc - y%gil%ref(y%gil%faxi) ) * y%gil%vres + y%gil%voff 
      nc(1:numchan) = (drange(1:numchan) - huv%gil%voff) / huv%gil%vres + huv%gil%ref(huv%gil%faxi)
    else if (mytype.eq.'FREQUENCY') then
      ! drange = nc - y%gil%ref(y%gil%faxi) ) * y%gil%vres + y%gil%voff 
      nc(1:numchan) = (drange(1:numchan) - huv%gil%freq) / huv%gil%fres + huv%gil%ref(huv%gil%faxi)
    else
      call map_message(seve%e,rname,"Type of value '"//trim(mytype)//"' not supported")
      error = .true.
      defmap_center = string
      return
    endif
    !
    do i=2,numchan,2
      if (nc(i).lt.nc(i-1)) then
        ntmp = nc(i)
        nc(i) = nc(i-1)
        nc(i-1) = ntmp
      endif
    enddo
    !
    call do_uv_average(rname,huv,hraw,numchan,nc,goodch,error)
    ! Copy to UVCALIB & UVSELF
    call gdf_copy_header(hraw,hcalib,error)
    call gdf_copy_header(hraw,hself,error)
    duvcalib(:,:) = duvraw(:,:)     ! A Physical copy
    duvself => duvcalib             ! A Pointer : no difference since only 1 channel
    call map_uvgildas ('UVSELF',hself,error,duvself)
  else
    !
    ! Here, we shall use the UV_COMPRESS approach 
    ! combined with the UV_CONTINUUM methodology
    call gdf_copy_header(huv,hraw,error)
    huv%r2d => duv
    ncomp = max(0,SELF_COMPRESS)
    call sub_cont_average(comm,huv,hraw,ncomp,error)
    ! !call gdf_print_header (hraw) 
    !
    ! Work with complete structure
    hraw%gil%dim(1) = 3*hraw%gil%nchan + 7 + hraw%gil%ntrail ! and nothing more ...
    if (hraw%gil%ntrail.ne.0) then
      do i=1,code_uvt_last
        if (hraw%gil%column_pointer(i).gt.hraw%gil%lcol) hraw%gil%column_pointer(i) =  &
          & hraw%gil%column_pointer(i) + hraw%gil%dim(1) - huv%gil%dim(1)
      enddo
    endif
    !
    ni = huv%gil%dim(1)
    nu = hraw%gil%dim(1)
    nv = hraw%gil%dim(2)
    nt = hraw%gil%ntrail
    nchan = hraw%gil%nchan
    !
    allocate(duvraw(hraw%gil%dim(1),huv%gil%dim(2)),stat=ier)
    if (ier.ne.0) then
      call map_message(seve%e,rname,'Self Calibration UV table allocation error')
      error = .true.
      return
    endif
    call compress_uv (duvraw,nu,nv,nchan,duv,ni,ncomp,nt,dchanflag)   
    !
    ! Copy this one to CALIB area
    call gdf_copy_header(hraw,hcalib,error)
    allocate(duvcalib(hraw%gil%dim(1),huv%gil%dim(2)),stat=ier)
    if (ier.ne.0) then
      call map_message(seve%e,rname,'CALIB Memory allocation error')
      error = .true.
      return
    endif
    duvcalib(:,:) = duvraw(:,:)
    call map_uvgildas('UVCALIB',hcalib,error,duvcalib)
    !
    call map_uvgildas('UVRAW',hraw,error,duvraw)
    !
    if (nchan.gt.1) then
      !
      ! The convert the table to a Continuum one
      channels = [1,nchan,1]
      call uv_cont_header(rname,hcalib,hself,channels,expand_cont,error)
      if (error) return
      ! and then the data
      allocate(duvself(hself%gil%dim(1),hself%gil%dim(2)),chflag(nchan),stat=ier) 
      if (ier.ne.0) then
        call map_message(seve%e,rname,'SELF Memory allocation error')
        error = .true.
        return
      endif
      chflag = 1
      ! 1=code_uvt, specifying UV Table shape
      hself%r2d => duvself
      hcalib%r2d => duvcalib
      call t_continuum(hcalib,hself,channels,chflag,1,error) ! No: ,spindex,ofreq)
      if (error) return
    else
      !
      ! Just point to that
      call gdf_copy_header(hcalib,hself,error)
      duvself => duvcalib
    endif
    call map_uvgildas ('UVSELF',hself,error,duvself)
  ENDIF
  !
  if (self_debug) then
    Print *,'EXPAND_CONT ',expand_cont,' SELF_COMPRESS ',self_compress
    read(*,*) expand_cont
  endif
  !
  ! At this stage, we should reset MAP_CENTER again...
  defmap_center = string
  !
  ! And force Weight re-computation
  do_weig = .true.
end subroutine init_selfcal
!
subroutine do_uv_average(rname,hin,hou,numchan,nc,goodch,error)
  use gildas_def
  use gkernel_interfaces
  use imager_interfaces, only : map_message, sub_uv_average
  use image_def
  use gbl_format
  use gbl_message
  !---------------------------------------------------------------------
  ! @ private
  !*
  ! IMAGER --  Support for command SELFCAL
  !
  !   Average all selected line channels to 1 continuum channel
  !   Driver routine: prepare header and call per-visibility processing
  !   subroutine
  !!
  !---------------------------------------------------------------------
  character(len=*), intent(in) :: rname     !! Caller
  type(gildas), intent(in) :: hin           !! Input raw visibility header
  type(gildas), intent(inout) :: hou        !! Output Self-Cal header
  integer, intent(in) :: numchan            !! Number of channel ranges
  integer, intent(in) :: nc(numchan)        !! Channel ranges
  integer, intent(in) :: goodch(:)          !! Selected channels
  logical, intent(out) :: error             !! Logical error flag
  !
  ! Local ---
  character(len=filename_length) :: text
  integer(4) :: nchan, nsum, i
  real(8) :: freq
  !
  ! Code ----
  error = .false.
  !
  if (hin%loca%size.eq.0 .or. .not.associated(hin%r2d)) then
    call map_message(seve%e,rname,'Input UV data is not allocated')
    error = .true.
    return
  endif
  !
  ! Define number of channels being added and effective mean channel
  nchan = 0
  nsum = 0
  do i=2,numchan,2
    nchan = nc(i)-nc(i-1)+1+nchan
    nsum = (nc(i)+nc(i-1))*(nc(i)-nc(i-1)+1)/2+nsum
    if (i.eq.2) then
      write (text,'(2(A,I0))') 'Averaging channels from ',   &
              nc(i-1),' to ',nc(i)
    else
      write (text,'(2(A,I0))') '              and  from ',   &
              nc(i-1),' to ',nc(i)
    endif
    call map_message(seve%i,rname,text)
  enddo
  !
  ! Define output image
  call gdf_copy_header (hin, hou, error)
  !
  ! Do not forget trailing columns
  hou%gil%dim(1) = 10+hou%gil%ntrail
  hou%gil%inc(1) = hou%gil%inc(1)*nchan
  hou%gil%ref(1) = 1.d0-(float(nsum)/nchan-hou%gil%ref(1))/nchan
  hou%gil%vres = nchan*hou%gil%vres
  hou%gil%fres = nchan*hou%gil%fres
  hou%gil%nchan = 1
  call gdf_uv_shift_columns(hin,hou)
  call gdf_setuv (hou, error)
  if (error) return
  !
  ! Bring reference channel to 1
  freq = (1.d0-hou%gil%ref(1))*hou%gil%inc(1) + hou%gil%val(1)
  hou%gil%ref(1) = 1.
  hou%gil%val(1) = freq
  ! Set Frequency accordingly
  hou%gil%freq = freq
  !
  call sub_uv_average (hou%r2d, hou%gil%dim(1),hou%gil%dim(2),   &
   &    hou%gil%nlead,  hou%gil%ntrail,               &
   &    hin%r2d ,hin%gil%dim(1),nc,numchan, goodch)
  hou%loca%size = hou%gil%dim(1) * hou%gil%dim(2)
end subroutine do_uv_average
!
subroutine sub_uv_average (out,nx,nv,nlead,ntrail,inp,ny,nc,num,goodch)
  use gildas_def
  !---------------------------------------------------------------------
  ! @ private
  !*
  ! IMAGER -- Support for command SELFCAL  
  !   Average all selected line channels to 1 continuum channel
  !!
  !---------------------------------------------------------------------
  integer(kind=index_length) :: nx    !! Size of output visibility
  integer(kind=index_length) :: nv    !! Number of visibilities
  integer, intent(in) :: nlead        !! Leading columns
  integer, intent(in) :: ntrail       !! Trailing columns
  real :: out(nx,nv)                  !! Output visibilities
  integer(kind=index_length) :: ny    !! Size of input visibility
  real :: inp(ny,nv)                  !! Input visibilities
  integer :: num                      !! Number of ranges times 2
  integer :: nc(num)                  !! Range boundaries
  integer :: goodch(:)                !! Channel flags
  !
  ! Local ---
  integer :: k,kk,l
  integer(kind=index_length) :: j
  real :: a,b,c
  !
  ! Code ----
  do j=1,nv
    out(1:nlead,j) = inp(1:nlead,j)
    a = 0.0
    b = 0.0
    c = 0.0
    do l=2,num,2
      do k=nc(l-1),nc(l)
        if (goodch(k).gt.0) then
          kk = nlead+3*k
          if (inp(kk,j).gt.0) then
            a = a+inp(kk-2,j)*inp(kk,j)
            b = b+inp(kk-1,j)*inp(kk,j)
            c = c+inp(kk  ,j)
          endif
        endif
      enddo
    enddo
    if (c.ne.0.0) then
      out(8,j) =a/c
      out(9,j) =b/c
      out(10,j)=c              ! time*band
    else
      out(8,j) =0.0
      out(9,j) =0.0
      out(10,j)=0.0
    endif
    if (ntrail.gt.0) out(11:nx,j) = inp(ny-ntrail+1:ny,j)
  enddo
end subroutine sub_uv_average
!
subroutine selfcal_comm(line,comm,error)
  use gkernel_interfaces
  !---------------------------------------------------------------------
  ! @ private
  !*
  ! IMAGER -- Support for command SELFCAL
  !
  !   This actually cannot be in the CLEAN\ language, because it implies
  !   re-entrancy of RUN_CLEAN - Not a big issue, but currently forbidden 
  !   by our coding practice, so we place it the ADVANCED\ language
  !!
  !---------------------------------------------------------------------
  character(len=*), intent(in) :: comm        !! Command name
  character(len=*), intent(inout) :: line     !! Command line
  logical, intent(inout) :: error             !! Logical error flag
  !
  ! Local ---
  character(len=256) :: argum
  integer :: na
  logical :: do_insert
  !
  ! Code ----
  do_insert = sic_lire().eq.0
  if (sic_present(1,0)) then
    call exec_program('@ p_selfcal /WIDGET')
  else
    if (sic_narg(0).eq.0) then
      call sic_get_char('SELF_MODE',argum,na,error)
    else
      ! Pass all arguments "as they are"
      na = index(line,trim(comm))+8
      argum = line(na:)
    endif
    call exec_program('@ p_selfcal '//argum)
  endif
  if (do_insert) call sic_insert_log(line)
end subroutine selfcal_comm
!
subroutine switch_uv_cont(line,comm,do_cct,code,htmp,error)
  use gildas_def 
  use clean_arrays
  use gbl_message
  use gkernel_interfaces
  use imager_interfaces, except_this => switch_uv_cont
  !---------------------------------------------------------------------
  ! @ private
  !
  ! IMAGER --  Support for command UV_MAP /CONT (and UV_SELF)
  !
  ! Temporarily point towards the "continuum" UV data
  !---------------------------------------------------------------------
  character(len=*), intent(inout) :: line     !! Command line
  character(len=*), intent(in)    :: comm     !! Command name
  character(len=2), intent(inout) :: code     !! UV buffer code
  type(gildas), intent(out) :: htmp           !! UV header
  logical, intent(in) :: do_cct     !! Remove Clean Components before ?
  logical, intent(inout) :: error   !! Logical error flag
  !
  ! Constants
  ! ' UV_MAP', '/TRUNCATE',  '/FIELDS', '/RANGE', '/CONTINUUM', '/INDEX'
  integer, parameter :: o_trunc=1
  integer, parameter :: o_field=2
  integer, parameter :: o_range=3
  integer, parameter :: o_cont=4
  integer, parameter :: o_index=5
  !
  ! Local ---
  integer :: channels(3)
  integer :: ier
  !
  ! Code ----
  call gildas_null(htmp,type='UVT')
  call uv_buffer_finduv(code)           ! Get next UV buffer code 
  call gdf_copy_header(huv,htmp,error)
  !
  ! Select the data set. Can be UV_DATA, UV_MODEL, or UV_RESIDUAL
  ! We may have an issue about data sorting here,
  ! as well as Weight re-computation
  call uvdata_select(comm,error)
  if (error) return
  !
  if (do_cct) then
    ! The UV data may have changed since last UV_MAP /CONT command
    if (huvc%loca%size.eq.0) then
      call map_message(seve%e,comm,'Continuum UV data not defined')
      error = .true.
      return
    endif
  else
    call sub_uvcont_header(line,error,huvc,channels,o_cont,o_index,o_range)
    if (error) return
    !
    if (allocated(duvc)) deallocate(duvc)
    allocate(duvc(huvc%gil%dim(1),huvc%gil%dim(2)),stat=ier)
    huvc%loca%size = huvc%gil%dim(1)*huvc%gil%dim(2)
    huvc%r2d => duvc
    huv%r2d => duv
    call sub_uvcont_data(line,huvc,huv,channels,dchanflag,o_index,error)
    if (error) return
    !
    ! Weights must be recomputed
    do_weig = .true.
    call sic_delvariable ('UVCONT',.false.,error)
    call sic_mapgildas('UVCONT',huvc,error,duvc)
  endif
  !
  ! Point to this UV data set
  call gdf_copy_header(huvc,huv,error)
  duv => duvc
  !
end subroutine switch_uv_cont
 
