subroutine clean_make(method,hclean,clean,tcc,gauss)
  use gbl_message
  use clean_def
  use image_def
  use gkernel_interfaces
  use imager_interfaces, only : mulgau, map_message
  !---------------------------------------------------------------------
  ! @ private-mandatory
  !*
  ! IMAGER -- Support for CLEAN
  !
  !   Build clean image from Component List of different sizes,
  ! using the Clean Component List structure.
  !!
  !---------------------------------------------------------------------
  type (clean_par), intent(in) :: method         !! Clean method parameters
  type (gildas), intent(inout) :: hclean         !! Clean header
  real, intent(inout) :: clean(hclean%gil%dim(1),hclean%gil%dim(2))  !! Clean data
  type (cct_par), intent(in) :: tcc(method%n_iter)    !! Clean component list
  real, intent(in), optional :: gauss(:)        !! Size of Gaussian sources
  !
  real(8), parameter :: pi=3.141592653589793d0
  !
  ! Local ---
  real, allocatable :: bsize(:)
  real :: dg
  integer :: ig,ng,kc
  integer nx,ny,nc,ix,iy,ic,imax,ndim,dim(2),ier
  real xinc, yinc,rmax,flux,r,fact,cmajor,cminor
  real, allocatable :: wfft(:)
  complex, allocatable :: ft(:,:)
  !
  ! Code ----
  nc = method%n_iter
  nx = hclean%gil%dim(1)
  ny = hclean%gil%dim(2)
  allocate (wfft(2*max(nx,ny)),ft(nx,ny),stat=ier)
  if (ier.ne.0) then
    call map_message(seve%e,'CLEAN','Memory allocation error in CLEAN_MAKE')
    return
  endif
  xinc = hclean%gil%convert(3,1)
  yinc = hclean%gil%convert(3,2)
  ndim = 2
  dim(1) = nx
  dim(2) = ny
  !
  if (method%method.eq.'SDI'.or.method%method.eq.'MULTI') then
    ft(:,:) = cmplx(clean,0.0)
    !
    !D Print *,'Calling FOURT'
    call fourt(ft,dim,ndim,-1,0,wfft)
    !
    ! Beam Area = PI * BMAJ * BMIN / (4 * LOG(2) ) for flux density
    ! normalisation
    fact = method%major*method%minor*pi/(4.0*log(2.0))   &
         &    /abs(xinc*yinc)/(nx*ny)
    !D Print *,'Calling MULGAU ',method%major, method%minor, method%angle
    call mulgau(ft,nx,ny,   &
         &    method%major,method%minor,method%angle,   &
         &    fact,xinc,yinc,-1)
    !D Print *,'Calling FOURT'
    call fourt(ft,dim,ndim,1,1,wfft)
    !
    ! Extract Real part
    !D Print *,'Done CLEAN'
    clean = real(ft)
    deallocate (wfft,ft)
  else
    if (present(gauss)) then
      ng = size(gauss)
      allocate (bsize(ng),stat=ier)
      bsize = gauss
    else
      ng = 1
      allocate (bsize(ng),stat=ier)
      bsize = 0
    endif
    ! !Print *,'NG ',ng,bsize*(180*3600/acos(-1.d0))
    !
    clean = 0.0
    do ig=1,ng
      dg =  bsize(ig)**2
      cmajor = sqrt(method%major**2+dg)
      cminor = sqrt(method%minor**2+dg)
      !
      ! Check minimum flux
      ft = 0.0
      kc = 0
      flux = 0
      !
      ! Convolve source list into residual map ---
      if (ng.gt.1) then
        do ic=1,nc
          if (tcc(ic)%size.eq.bsize(ig)) then
            ! !Print *,'IC ',ic,tcc(ic)%ix,tcc(ic)%iy
            ix = tcc(ic)%ix
            iy = tcc(ic)%iy
            ft(ix,iy) = ft(ix,iy) + tcc(ic)%value
            kc = kc+1
          endif
        enddo
      else
        do ic=1,nc
          ! !Print *,'IC ',ic,tcc(ic)%ix,tcc(ic)%iy
          ix = tcc(ic)%ix
          iy = tcc(ic)%iy
          ft(ix,iy) = ft(ix,iy) + tcc(ic)%value
        enddo
        kc = nc
      endif
      ! !Print *,'Size ',ig,' Counter ',kc,' Flux ',sum(real(ft))
      !
      ndim = 2
      dim(1) = nx
      dim(2) = ny
      !D Print *,'Calling FOURT'
      call fourt(ft,dim,ndim,-1,0,wfft)
      !
      ! Beam Area = PI * BMAJ * BMIN / (4 * LOG(2) ) for flux density
      ! normalisation
      fact = method%major*method%minor*pi/(4.0*log(2.0))   &
           &    /abs(xinc*yinc)/(nx*ny)
      if (fact.eq.0) fact = 1.0 /(nx*ny) ! Used when calling clean_make to Restore a Mosaic
      !D Print *,'Calling MULGAU ',method%major, method%minor, method%angle
      call mulgau(ft,nx,ny,   &
           &    cmajor,cminor,method%angle,   &
           &    fact,xinc,yinc,-1)
      !D Print *,'Calling FOURT'
      call fourt(ft,dim,ndim,1,1,wfft)
      !
      ! Extract Real part
      !D Print *,'Done CLEAN'
      clean = clean+real(ft)
    enddo
  endif
end subroutine clean_make
!
subroutine clean_make_cct(method,hclean,clean,dcct,gauss)
  use gbl_message
  use clean_def
  use image_def
  use gkernel_interfaces
  use imager_interfaces, only : mulgau, map_message
  !---------------------------------------------------------------------
  ! @ private-mandatory
  !*
  ! IMAGER  -- Support for command CLEAN
  !
  !   Build clean image from Component List of different sizes, using
  !   the 3-D component array.
  !!
  !---------------------------------------------------------------------
  type (clean_par), intent(in) :: method         !! Clean method parameters
  type (gildas), intent(in)    :: hclean         !! Clean header
  real, intent(inout) :: clean(hclean%gil%dim(1),hclean%gil%dim(2))  !! Clean array
  real, intent(in) :: dcct(:,:,:)                !! Clean components
  real, intent(in), optional :: gauss(:)         !! Gaussian sizes
  !
  real(8), parameter :: pi=3.141592653589793d0
  !
  ! Local ---
  real, allocatable :: bsize(:)
  real :: dg
  integer :: ig,ng,kc
  integer nx,ny,nc,ix,iy,ic,imax,ndim,dim(2),ier
  real xinc, yinc,rmax,flux,r,fact,cmajor,cminor
  real, allocatable :: wfft(:)
  complex, allocatable :: ft(:,:)
  integer :: iplane
  real(8) :: xref,xval, yref, yval
  !
  ! Code ----
  nc = method%n_iter
  if (nc.eq.0) nc = size(dcct,2)
  nx = hclean%gil%dim(1)
  ny = hclean%gil%dim(2)
  allocate (wfft(2*max(nx,ny)),ft(nx,ny),stat=ier)
  if (ier.ne.0) then
    call map_message(seve%e,'CLEAN','Memory allocation error in Clean_Make90')
    return
  endif
  xref = hclean%gil%convert(1,1)
  xval = hclean%gil%convert(2,1)
  xinc = hclean%gil%convert(3,1)
  yref = hclean%gil%convert(1,2)
  yval = hclean%gil%convert(2,2)
  yinc = hclean%gil%convert(3,2)
  ndim = 2
  dim(1) = nx
  dim(2) = ny
  !
  if (method%method.eq.'SDI'.or.method%method.eq.'MULTI') then
    ft(:,:) = cmplx(clean,0.0)
    !
    !D Print *,'Calling FOURT'
    call fourt(ft,dim,ndim,-1,0,wfft)
    !
    ! Beam Area = PI * BMAJ * BMIN / (4 * LOG(2) ) for flux density
    ! normalisation
    fact = method%major*method%minor*pi/(4.0*log(2.0))   &
         &    /abs(xinc*yinc)/(nx*ny)
    !D Print *,'Calling MULGAU ',method%major, method%minor, method%angle
    call mulgau(ft,nx,ny,   &
         &    method%major,method%minor,method%angle,   &
         &    fact,xinc,yinc,-1)
    !D Print *,'Calling FOURT'
    call fourt(ft,dim,ndim,1,1,wfft)
    !
    ! Extract Real part
    !D Print *,'Done CLEAN'
    clean = real(ft)
    deallocate (wfft,ft)
  else
    iplane = 1 ! Called by plane
    if (present(gauss)) then
      ng = size(gauss)
      allocate (bsize(ng),stat=ier)
      bsize = gauss
    else
      ng = 1
      allocate (bsize(ng),stat=ier)
      bsize = 0
    endif
    do ig = ng,1,-1
      if (bsize(ig).ne.0) then
        ng = ig
        exit
      endif
    enddo
    bsize = bsize*acos(-1.)/180/3600 ! To be changed
    ! !Print *,'NG ',ng,bsize*(180*3600/acos(-1.d0))
    !
    clean = 0.0
    do ig=1,ng
      !
      ! Check minimum flux
      ft = 0.0
      kc = 0
      flux = 0
      !
      ! Convolve source list into residual map ---
      do ic=1,nc
        if (dcct(4,ic,iplane).eq.bsize(ig)) then
          ! ! Print *,'IC ',ic,dcct(1:4,ic,iplane)
          if (dcct(3,ic,iplane).eq.0) exit
          !!Print *,'IC ',ic,tcc(ic)%ix,tcc(ic)%iy
          ix = nint( (dcct(1,ic,iplane)-xval) / xinc + xref )
          iy = nint( (dcct(2,ic,iplane)-yval) / yinc + yref )
          ft(ix,iy) = ft(ix,iy) + dcct(3,ic,iplane)
          kc = kc+1
        endif
      enddo
      ! !Print *,'Size ',ig,' Counter ',kc,' Flux ',sum(real(ft))
      !
      if (method%major.ne.0) then
        dg =  bsize(ig)**2
        cmajor = sqrt(method%major**2+dg)
        cminor = sqrt(method%minor**2+dg)
      else
        cmajor = bsize(ig)
        cminor = bsize(ig)
      endif
      !
      if (cmajor.gt.0.) then
        ndim = 2
        dim(1) = nx
        dim(2) = ny
        !D Print *,'Calling FOURT'
        call fourt(ft,dim,ndim,-1,0,wfft)
        !
        ! Beam Area = PI * BMAJ * BMIN / (4 * LOG(2) ) for flux density normalisation
        fact = method%major*method%minor*pi/(4.0*log(2.0))   &
             &    /abs(xinc*yinc)/(nx*ny)
        if (fact.eq.0) fact = 1.0/(nx*ny) ! Used when calling clean_make_cct to Restore a Mosaic
        !D Print *,'Calling MULGAU ',method%major, method%minor, method%angle
        call mulgau(ft,nx,ny,   &
             &    cmajor,cminor,method%angle,   &
             &    fact,xinc,yinc,-1)
        !D Print *,'Calling FOURT'
        call fourt(ft,dim,ndim,1,1,wfft)
      endif
      !
      ! Extract Real part
      !D Print *,'Done CLEAN'
      clean = clean+real(ft)
    enddo
  endif
end subroutine clean_make_cct
!
subroutine check_area(method,head,quiet)
  use clean_def
  use clean_default
  use image_def
  use gbl_message
  use imager_interfaces, only : map_message, check_box
  !---------------------------------------------------------------------
  ! @ private
  !*
  ! IMAGER -- Support routine for CLEAN
  !
  !   Check search area for Cleaning, and guess the maximum number of 
  !   iterations
  !!
  !---------------------------------------------------------------------
  type (gildas), intent(in) :: head             !! Image header
  type (clean_par), intent(inout) :: method     !! Cleaning parameters  
  logical, intent(in) :: quiet                  !! Printout level
  !
  character(len=5) :: rname = 'CLEAN'
  character(len=16) :: citer = 'ITERATIONS'
  character(len=16) :: ccomp = 'COMPONENTS'
  character(len=16) :: cnois = 'NOISE'
  character(len=16) :: csigm = 'SIGMA'
  character(len=16) :: cstop = 'STOP'
  !
  ! Local ---
  character(len=16) :: ctype
  real ares,hmax
  integer max_iter0, nx, ny
  character(len=message_length) :: chain
  integer :: ier, nc, kc
  real :: scale, factor, anoise
  logical :: error
  !
  ! Code ----
  nx = head%gil%dim(1)
  ny = head%gil%dim(2)
  ! get BLC & TRC
  call check_box(nx,ny,method%blc,method%trc)
  !
  ! Support for Clean
  method%box = (/method%blc(1),method%blc(2),   &
       &    method%trc(1),method%trc(2)/)
  !
  ier = -1
  hmax = max(abs(head%gil%rmax),abs(head%gil%rmin))
  anoise = head%gil%noise
  if (head%gil%rms.gt.0) then
    if (head%gil%rms.lt.head%gil%noise) then
      call map_message(seve%w,rname, &
      & 'RMS noise of DIRTY image is lower than theoretical noise')
      anoise = head%gil%rms
    endif
  endif
  if (clean_stop(1).ne." ") then
    read(clean_stop(1),*,iostat=ier) factor
  endif  
  !
  if (ier.eq.0) then
    ! Default stopping always based on Noise
    ares = anoise
    !
    ctype = clean_stop(2)
    call sic_upper(ctype)
    if (ctype.eq.'S') then
      if (.not.quiet) call map_message(seve%w,rname, &
        & 'CLEAN_STOP[2] is ambiguous, using SIGMA, not STOP')
      clean_stop(2) = 'Sigma'
      ctype = 'SIGMA'
      nc = 5
    else
      nc = len_trim(ctype)
    endif
    !
    scale = 0
    if ((ctype.eq.cnois(1:nc)) .or. &
      & (ctype.eq.csigm(1:nc)) ) then
      scale = anoise 
      ! ARES and FRES have no meaning then
      user_method%ares = 0
      user_method%fres = 0
    else if (ctype.eq.'%') then
      scale = hmax * 0.01
      user_method%fres = factor*0.01  ! Make it remanent
      user_method%ares = 0
    else if (ctype(nc-1:nc).eq.'JY') then
      kc = nc-2
      if (kc.gt.0) then
        select case (ctype(1:kc))
        case ('MILLI','M')
          scale = 1E-3
        case ('MICRO')
          scale = 1E-6
        case ('KILO','K')
          scale = 1E3
        case ('MEGA')
          scale = 1E6
        case default
          scale = 0
        end select
      else
        scale = 1.0
      endif
      user_method%fres = 0
    else if ( (ctype.eq.citer(1:nc)) .or. &
      & (ctype.eq.ccomp(1:nc)) ) then
      method%m_iter = max(0,nint(factor))
      scale = anoise
      if (method%m_iter.ne.0) then
        factor = 0.
        user_method%m_iter = method%m_iter    ! Make it remanent
      else
        factor = 1.
      endif
    else if (ctype.eq.cstop(1:nc)) then
      method%converge = max(0,nint(factor))
      scale = anoise
      if (method%converge.ne.0) then
        factor = 0.
      else
        factor = 1.
      endif
      user_method%converge = method%converge  ! Make it remanent
      clean_stop = " "                        ! Reset CLEAN_STOP string
    endif
    !
    if (scale.ne.0) then
      ares = factor*scale
      if (user_method%fres.eq.0) user_method%ares = ares
    else
      ier = 1
    endif
  endif
  !
  if (ier.gt.0) then
    if (.not.quiet) call map_message(seve%w,rname, &
    & 'Invalid CLEAN_STOP, using ARES, FRES or NITER')
  endif
  if (ier.ne.0) then
    ares = method%ares
    if (ares.eq.0) ares = method%fres*hmax
    if (ares.eq.0) ares = anoise
  else if (.not.quiet) then
    call map_message(seve%i,rname,'Using CLEAN_STOP '// &
      & trim(clean_stop(1))//' '//trim(clean_stop(2)))
  endif
  !
  if (.not.quiet) then
    write(chain,'(A,1pg11.3)') 'Cleaning down to ',ares
    call map_message(seve%i,rname,chain)
  endif
  !
  if (hmax.lt.ares) then
    max_iter0 = 0
    if (.not.quiet) then
      write(chain,'(A,1pg11.3,A)') 'Peak flux ',hmax,' is smaller, no Cleaning needed'
      call map_message(seve%w,rname,chain)
    endif
  else if (ares.ne.0) then
    max_iter0 = log(ares/hmax)/log(1.0-method%gain)*  &
         &(method%trc(1)-method%blc(1)+1)*(method%trc(2)-method%blc(2)+1)/50.0
    if (max_iter0.lt.0) then
      chain = 'Unable to guess Iteration number - set to 100 '
      call map_message(seve%w,rname,chain)
      write(chain,*) 'CLEAN_Ares ',ares,'  Hmax',hmax,'  CLEAN_Gain ',method%gain
      call map_message(seve%w,rname,chain)
      max_iter0 = 100
    endif
    !
    if (.not.quiet) then
      if (ares.lt.head%gil%noise) then
        write(chain,'(A,1pg11.3)') 'Expected noise is larger... ',head%gil%noise
        call map_message(seve%w,rname,chain)
      endif
      write(chain,'(A,i12,a)') 'May need ',max_iter0,' iterations'
      call map_message(seve%i,rname,chain)
    endif
  endif
  !
  if (method%m_iter.eq.0) then
    if (.not.cct_dynamic) method%m_iter = max_iter0
    method%ares = ares
  endif
end subroutine check_area
!
subroutine beam_plane(method,hbeam,hdirty)
  use clean_def
  use image_def
  use gbl_message
  use imager_interfaces, only : map_message, beam_for_channel
  !---------------------------------------------------------------------
  ! @ private
  !*
  ! IMAGER  -- Support routine for CLEAN
  !
  !   Define beam plane and check if fit is required
  !!
  !---------------------------------------------------------------------
  type (clean_par), intent(inout) :: method     !! Clean method parameters
  type (gildas), intent(in) :: hbeam            !! Beam header
  type (gildas), intent(in) :: hdirty           !! Used to figure out which plane
  !
  integer :: ibeam
  !
  ! Check if beam parameters must be recomputed
  if (hbeam%gil%dim(4).le.1) then
    method%ibeam = 1
  else
    ! Cannot set method%ibeam  directly
    ibeam = beam_for_channel(method%iplane,hdirty,hbeam)
    method%ibeam = ibeam
  endif
end subroutine beam_plane
!
function beam_for_channel(iplane,hdirty,hbeam)
  use image_def
  !---------------------------------------------------------------------
  ! @ public
  !*
  ! IMAGER  -- Support routine for CLEAN
  !
  !   Find out the appropriate beam channel for an image channel
  !!
  !---------------------------------------------------------------------  
  integer :: beam_for_channel !! intent(out)    !! Beam channel
  type(gildas), intent(in) :: hdirty            !! Dirty header
  type(gildas), intent(in) :: hbeam             !! Beam header
  integer, intent(in) :: iplane                 !! Dirty channel number
  !
  ! Local ---
  real(8) :: i_freq
  integer :: ibeam, nstep, v_axis
  !
  ! Code ----
  if (hbeam%gil%faxi.eq.0) then
    v_axis = 3
  else
    v_axis = hbeam%gil%faxi  ! Which axis is the Velocity Axis ?
  endif
  !
  if (hbeam%gil%dim(v_axis).le.1) then
    beam_for_channel = 1
  else
    nstep = hdirty%gil%dim(3) / hbeam%gil%dim(v_axis)
    if (hbeam%gil%dim(v_axis)*nstep.ne.hdirty%gil%dim(3)) nstep = nstep+1
    ! ibeam = (iplane+nstep-1)/nstep  ! Approximate
    ! This is the better way of doing it:
    i_freq = (iplane-hdirty%gil%ref(3))*hdirty%gil%fres + hdirty%gil%freq
    ibeam = nint((i_freq-hbeam%gil%freq)/hbeam%gil%fres + hbeam%gil%ref(v_axis))
    ibeam = min(max(1,ibeam),hbeam%gil%dim(v_axis)) ! Just  in case
    !
    beam_for_channel = ibeam
  endif
end function beam_for_channel
!
subroutine old_clean_make90(method,hclean,clean,tcc)
  use gbl_message
  use clean_def
  use image_def
  use gkernel_interfaces
  use imager_interfaces, only : mulgau, map_message
  !---------------------------------------------------------------------
  ! @ private
  !
  ! IMAGER
  !   Build clean image from Component List
  !---------------------------------------------------------------------
  type (clean_par), intent(inout) :: method    ! Clean method parameters
  type (gildas), intent(inout) :: hclean       ! Clean header
  real, intent(inout) :: clean(hclean%gil%dim(1),hclean%gil%dim(2))
  type (cct_par), intent(in) :: tcc(method%n_iter)  ! Clean components
  !
  real(8), parameter :: pi=3.141592653589793d0
  !
  integer nx,ny,nc,ix,iy,ic,imax,ndim,dim(2),ier
  real xinc, yinc,rmax,flux,r,fact
  real, allocatable :: wfft(:)
  complex, allocatable :: ft(:,:)
  !
  nc = method%n_iter
  nx = hclean%gil%dim(1)
  ny = hclean%gil%dim(2)
  allocate (wfft(2*max(nx,ny)),ft(nx,ny),stat=ier)
  if (ier.ne.0) then
    call map_message(seve%e,'CLEAN','Memory allocation error in Clean_Make90')
    return
  endif
  xinc = hclean%gil%convert(3,1)
  yinc = hclean%gil%convert(3,2)
  !
  if (method%method.eq.'SDI'.or.method%method.eq.'MULTI') then
    ft(:,:) = cmplx(clean,0.0)
  else
    !
    ! Check minimum flux
    ft = 0.0
    !
    if (.false.) then
      rmax = abs(tcc(1)%value)
      imax = 1
      do ic=1,nc
        r = abs(tcc(ic)%value)
        if (r.lt.rmax) then
          imax = ic
          rmax = r
        endif
      enddo
      if (imax.lt.nc-10) then
      !            WRITE(CHAIN,*)
      !     $      'I-CLEAN,  Truncating CLEAN component to '
      !     $      ,IMAX,' / ',NC
      !            CALL GAGOUT(CHAIN)
      !            NC = IMAX
      endif
    endif
    !
    ! Convolve source list into residual map ---
    if (method%bshift(3).eq.0) then
      do ic=1,nc
        ! !Print *,'IC ',ic,tcc(ic)%ix,tcc(ic)%iy
        ix = tcc(ic)%ix
        iy = tcc(ic)%iy
        ft(ix,iy) = ft(ix,iy) + tcc(ic)%value
      enddo
    else
      do ic=1,nc
        flux = 0.5*tcc(ic)%value
        ix = tcc(ic)%ix
        iy = tcc(ic)%iy
        ft(ix,iy) = ft(ix,iy) + flux
        ix = ix+method%bshift(1)
        iy = iy+method%bshift(2)
        ft(ix,iy) = ft(ix,iy) + flux
      enddo
    endif
  endif
  !
  ndim = 2
  dim(1) = nx
  dim(2) = ny
  !D Print *,'Calling FOURT'
  call fourt(ft,dim,ndim,-1,0,wfft)
  !
  ! Beam Area = PI * BMAJ * BMIN / (4 * LOG(2) ) for flux density
  ! normalisation
  fact = method%major*method%minor*pi/(4.0*log(2.0))   &
       &    /abs(xinc*yinc)/(nx*ny)
  !D Print *,'Calling MULGAU ',method%major, method%minor, method%angle
  call mulgau(ft,nx,ny,   &
       &    method%major,method%minor,method%angle,   &
       &    fact,xinc,yinc,-1)
  !D Print *,'Calling FOURT'
  call fourt(ft,dim,ndim,1,1,wfft)
  !
  ! Extract Real part
  !D Print *,'Done CLEAN'
  clean = real(ft)
  deallocate (wfft,ft)
end subroutine old_clean_make90
!
subroutine plunge_real (r,nx,ny,c,mx,my)
  !------------------------------------------------------------------------------
  ! @ private
  !*
  ! IMAGER  -- Support for UV_RESTORE and others
  !
  !   Plunge a Real array into a Complex array, at corners
  !   for subsequent FFTs
  !!
  !-----------------------------------------------------------------
  integer, intent(in)  :: nx,ny                !! Size of input array
  real, intent(in)     :: r(nx,ny)             !! Input real array
  integer, intent(in)  :: mx,my                !! Size of output array
  complex, intent(out) :: c(mx,my)             !! Output complex array
  !
  ! Local ---
  integer :: kx,ky,lx,ly,i,j
  !
  kx = nx/2+1
  lx = mx/2+1
  ky = ny/2+1
  ly = my/2+1
  !
  c = 0.0
  do j=1,ny
    do i=1,nx
      c(i-kx+lx,j-ky+ly) = cmplx(r(i,j),0.0)
    enddo
  enddo
end subroutine plunge_real
!
subroutine recent(nx,ny,z)
  !------------------------------------------------------------------------------
  ! @ private
  !*
  ! IMAGER  -- Support for UV_RESTORE and others
  !
  !   Recenters the Fourier Transform, for easier display. 
  !   The present version will only work for even dimensions.
  !!
  !------------------------------------------------------------------------------
  integer, intent(in) :: nx  !! X size
  integer, intent(in) :: ny  !! Y size
  complex, intent(inout) :: z(nx,ny)  !! Complex  array
  !
  ! Local ---
  complex :: tmp
  integer :: i, j
  !
  ! Code ----
  do j=1,ny/2
    do i=1,nx/2
      tmp = z(i+nx/2,j+ny/2)
      z(i+nx/2,j+ny/2) = z(i,j)
      z(i,j) = tmp
    enddo
  enddo
  !
  do j=1,ny/2
    do i=1,nx/2
      tmp = z(i,j+ny/2)
      z(i,j+ny/2) = z(i+nx/2,j)
      z(i+nx/2,j) = tmp
    enddo
  enddo
  !
  do j=1,ny
    do i=1,nx
      if (mod(i+j,2).ne.0) then
        z(i,j) = -z(i,j)
      endif
    enddo
  enddo
end subroutine recent
!
subroutine plunge_r4 (r,nx,ny,o,mx,my)
  !-----------------------------------------------------------------
  ! @ private
  !*
  ! IMAGER  -- Generic tool
  !     Plunge a Real array into another Array (smaller or bigger)
  !!
  !-----------------------------------------------------------------
  integer, intent(in)  :: nx,ny                !! Size of input array
  real, intent(in)     :: r(nx,ny)             !! Input real array
  integer, intent(in)  :: mx,my                !! Size of output array
  real, intent(out)    :: o(mx,my)             !! Output real array
  !
  ! Local ---
  integer im,ip,jm,jp,km,kp,lm,lp
  !
  ! Match the centers
  call pixel_range((mx-nx)/2,im,ip,km,kp,nx,mx)
  call pixel_range((mx-nx)/2,jm,jp,lm,lp,ny,my)
  !
  o = 0.
  o(im:ip,jm:jp) = r(km:kp,lm:lp)
end subroutine plunge_r4
